@@ -2,7 +2,7 @@
 
 /**
   * SquirrelMail Server Settings Backend Plugin
-  * Copyright (c) 2008-2009 Paul Lesniewski <paul@squirrelmail.org>
+  * Copyright (c) 2008-2010 Paul Lesniewski <paul@squirrelmail.org>
   * Licensed under the GNU GPL. For full terms see the file COPYING.
   *
   * @package plugins
@@ -78,7 +78,7 @@
   *               contain each setting value and its offset in the target
   *               file (in that order) if the setting is found more than
   *               once in the file (the caller should check if the return
-  *               value's first array element is an array itslef to
+  *               value's first array element is an array itself to
   *               determine if the result is non-scalar), or NULL if the
   *               setting is not found in the file or if an error occurred
   *               (and $quiet is TRUE).  Note that if $return_file_contents
@@ -97,140 +97,47 @@ function retrieve_server_setting_from_ftp($FTP, $host, $port, $use_ssl, $passive
 {
 
    $error = '';
-   $close_connection = FALSE;
-   $cwd = '.';
-   if ($remote_file[strlen($remote_file) - 1] == '/'
-    || $remote_file[strlen($remote_file) - 1] == '\\')
-      $remote_file = substr($remote_file, 0, -1);
+
+   $file = ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode,
+                        $user, $pass, $directory, $remote_file,
+                        $treat_as_empty_when_not_found, $quiet);
 
 
-   // connect to FTP server
+   // first deal with file-not-found
    //
-   if (!is_resource($FTP))
-   {
-      $FTP = ssb_ftp_connect($host, $port, $use_ssl, $passive,
-                             $user, $pass, $directory, $quiet);
-      $close_connection = TRUE;
-
-      // don't do anything if not connected
-      //
-      if (is_null($FTP))
-         return NULL;
-   }
-
-
-   // if already connected, note directory we started in
-   // and then make sure we're in the right directory
-   //
-   else if (!($cwd = ftp_pwd($FTP)) || (!empty($directory) && !sq_call_function_suppress_errors('ftp_chdir', array($FTP, $directory))))
-   {
-      if ($quiet) return TRUE;
-      sq_change_text_domain('server_settings_backend');
-
-      // show connect details when in debug mode
-      //
-      global $sm_debug_mode, $server_settings_backend_debug;
-      server_settings_backend_init();
-      if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
-      if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
-         $error = sprintf(_("Failed to change to directory \"%s\" on FTP host \"%s\""), $directory, $host);
-      else
-         $error = _("Unable to change directories on FTP server");
-
-      sq_change_text_domain('squirrelmail');
-   }
-
-
-   // override code above that might ask for closing
-   // the connection when this function exits...
-   // we want to cache all our connections, and closing
-   // a connection will b0rk someone else who wants to
-   // use the same connection
-   //
-   $close_connection = FALSE;
-
-
-   if (empty($error))
+   if (is_null($file))
    {
 
-      // get a temp file for downloading the target file
-      //
-      $temp = ssb_get_temp_file($quiet);
-      if (is_null($temp)) // implied $quiet is TRUE
-      {
-         if ($close_connection) ftp_close($FTP);
-         else @ftp_chdir($cwd);
-         return NULL;
-      }
-      list($FILE, $file_path) = $temp;
-
-
-      // now, actually pull file
-      //
-      if ($mode == 'ASCII') $mode = FTP_ASCII;
-      else $mode = FTP_BINARY;
-
-
-      // don't want to dump errors anywhere when missing file
-      // means nothing to us
-      //
       if ($treat_as_empty_when_not_found)
       {
-
-         if (!@ftp_fget($FTP, $FILE, $remote_file, $mode))
-         {
-            // parse_value() found in functions.php
-            //
-            $ret = parse_value('', $parse_pattern, $pattern_group_number, $quiet);
-            if ($return_file_contents)
-               $ret = array($ret, '');
-            if ($close_connection) ftp_close($FTP);
-            else @ftp_chdir($cwd);
-            return $ret;
-         }
-
+         // parse_value() found in functions.php
+         //
+         $ret = parse_value('', $parse_pattern, $pattern_group_number, $quiet);
+         if ($return_file_contents)
+            $ret = array($ret, '');
+         return $ret;
       }
-      else
-      {
 
-         if (!sq_call_function_suppress_errors('ftp_fget', array($FTP, $FILE, $remote_file, $mode)))
-         {
-            if ($quiet)
-            {
-               if ($close_connection) ftp_close($FTP);
-               else @ftp_chdir($cwd);
-               return NULL;
-            }
-            sq_change_text_domain('server_settings_backend');
-            $error = _("FTP error when attempting to retrieve a setting from the server");
-            sq_change_text_domain('squirrelmail');
-            fclose($FILE);
-            unlink($file_path);
-         }
+      if ($quiet) return NULL;
 
-      }
+      sq_change_text_domain('server_settings_backend');
+      $error = _("FTP error when attempting to retrieve a setting from the server");
+      sq_change_text_domain('squirrelmail');
 
    }
 
 
    // parse out our target value from the file
    //
-   if (empty($error))
+   else
    {
 
-      fclose($FILE);
-
-
-      // now parse out the value(s) from the file
       // parse_value_from_file() found in functions.php
       //
-      $ret = parse_value_from_file($file_path, $parse_pattern,
+      $ret = parse_value_from_file($file, $parse_pattern,
                                    $pattern_group_number, $quiet,
                                    $return_file_contents);
-      unlink($file_path);
-
-      if ($close_connection) ftp_close($FTP);
-      else @ftp_chdir($cwd);
+      unlink($file);
       return $ret;
 
    }
@@ -246,7 +153,6 @@ function retrieve_server_setting_from_ftp($FTP, $host, $port, $use_ssl, $passive
       global $oTemplate;
       $oTemplate->display('footer.tpl');
    }
-   @ftp_close($FTP);
    exit;
 
 }
@@ -338,7 +244,7 @@ function ftp_put_file($FTP, $file_contents, $allow_overwrite, $host,
    //
    else if (!($cwd = ftp_pwd($FTP)) || (!empty($directory) && !sq_call_function_suppress_errors('ftp_chdir', array($FTP, $directory))))
    {
-      if ($quiet) return TRUE;
+      if ($quiet) return NULL;
       sq_change_text_domain('server_settings_backend');
 
       // show connect details when in debug mode
@@ -407,7 +313,7 @@ function ftp_put_file($FTP, $file_contents, $allow_overwrite, $host,
          // and an error occured, so if it does, just return NULL
          //
          if (!ssb_ftp_delete($FTP, $host, $port, $use_ssl, $passive, $user,
-                           $pass, $directory, $remote_file, FALSE, $quiet))
+                           $pass, $directory, $remote_file, FALSE, FALSE, $quiet))
          {
             if ($close_connection) ftp_close($FTP);
             else @ftp_chdir($cwd);
@@ -434,7 +340,7 @@ function ftp_put_file($FTP, $file_contents, $allow_overwrite, $host,
          else @ftp_chdir($cwd);
          return NULL;
       }
-      list($FILE, $file_path) = $temp;
+      list($file_path, $FILE) = $temp;
 
 
       // now, write the file contents out to the local file
@@ -568,11 +474,7 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
    if ($remote_file[strlen($remote_file) - 1] == '/'
     || $remote_file[strlen($remote_file) - 1] == '\\')
       $remote_file = substr($remote_file, 0, -1);
-//TODO: path separator below may not work under all O/Ses(?)
-   $full_path = (empty($directory) ? $remote_file : $directory . '/' . $remote_file);
-   $path_directory = dirname($full_path);
-   $path_file = basename($full_path);
-
+      
 
    // connect to FTP server
    //
@@ -594,7 +496,7 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
    //
    else if (!($cwd = ftp_pwd($FTP)) || (!empty($directory) && !sq_call_function_suppress_errors('ftp_chdir', array($FTP, $directory))))
    {
-      if ($quiet) return TRUE;
+      if ($quiet) return NULL;
       sq_change_text_domain('server_settings_backend');
 
       // show connect details when in debug mode
@@ -620,37 +522,6 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
    $close_connection = FALSE;
 
 
-   // check if the file exists in the first place...
-   // if not, decide what to do
-   //
-   if (empty($error)
-    && !ftp_path_exists($FTP, $host, $port, $use_ssl, $passive, $user,
-                        $pass, $full_path, FALSE, $quiet))
-   {
-      if ($quiet || $ignore_non_existent)
-      {
-         if ($close_connection) ftp_close($FTP);
-         else @ftp_chdir($cwd);
-         return NULL;
-      }
-      sq_change_text_domain('server_settings_backend');
-      
-      // show connect details when in debug mode
-      //
-      global $sm_debug_mode, $server_settings_backend_debug;
-      server_settings_backend_init();
-      if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
-      if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
-         $error = sprintf(_("File \"%s\" not found in directory \"%s\" on FTP host \"%s\""), $path_file, $path_directory, $host);
-      else
-         $error = _("File not found");
-   
-      sq_change_text_domain('squirrelmail');
-   }
-
-
-   // now, actually get the file
-   //
    if (empty($error))
    {
 
@@ -663,7 +534,7 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
          else @ftp_chdir($cwd);
          return NULL;
       }
-      list($FILE, $file_path) = $temp;
+      list($file_path, $FILE) = $temp;
 
 
       // now, actually pull file
@@ -672,48 +543,61 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
       else $mode = FTP_BINARY;
 
 
-      // file has already been verified to exist,
-      // so errors here should be more serious
+      // don't want to dump errors anywhere when missing file
+      // means nothing to us
       //
-      if (!sq_call_function_suppress_errors('ftp_fget', array($FTP, $FILE, $remote_file, $mode)))
+      if ($ignore_non_existent || $quiet)
       {
-         if ($quiet)
+
+         if (!@ftp_fget($FTP, $FILE, $remote_file, $mode))
          {
             if ($close_connection) ftp_close($FTP);
             else @ftp_chdir($cwd);
             return NULL;
          }
-         sq_change_text_domain('server_settings_backend');
-      
-         // show connect details when in debug mode
-         //
-         global $sm_debug_mode, $server_settings_backend_debug;
-         server_settings_backend_init();
-         if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
-         if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
-            $error = sprintf(_("Could not retrieve file \"%s\" in directory \"%s\" on FTP host \"%s\""), $remote_file, $directory, $host);
-         else
-            $error = _("FTP error when attempting to retrieve file");
 
-         sq_change_text_domain('squirrelmail');
-         fclose($FILE);
-         unlink($file_path);
       }
-   
-
-      // success
-      //
-      else 
+      else
       {
-         if ($close_connection) ftp_close($FTP);
-         else @ftp_chdir($cwd);
-         fclose($FILE);
-         return $file_path;
+
+         if (!sq_call_function_suppress_errors('ftp_fget', array($FTP, $FILE, $remote_file, $mode)))
+         {
+
+            // show connect details when in debug mode
+            //
+            global $sm_debug_mode, $server_settings_backend_debug;
+            server_settings_backend_init();
+            if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+            if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+               $error = sprintf(_("Could not retrieve file \"%s\" in directory \"%s\" on FTP host \"%s\""), $remote_file, $directory, $host);
+            else
+               $error = _("FTP error when attempting to retrieve file");
+
+            sq_change_text_domain('squirrelmail');
+
+            if ($close_connection) ftp_close($FTP);
+            else @ftp_chdir($cwd);
+            fclose($FILE);
+            unlink($file_path);
+
+         }
+
       }
 
    }
 
 
+   // success
+   //
+   if (empty($error))
+   {
+      if ($close_connection) ftp_close($FTP);
+      else @ftp_chdir($cwd);
+      fclose($FILE);
+      return $file_path;
+   }
+
+
    // spew error
    //
    global $color;
@@ -724,7 +608,6 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
       global $oTemplate;
       $oTemplate->display('footer.tpl');
    }
-   @ftp_close($FTP);
    exit;
 
 }
@@ -769,6 +652,11 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
   *                               inspecting the target if at all possible,
   *                               but if not supported by the FTP server,
   *                               the value passed in here will be used.
+  * @param boolean $autodetect    When FALSE, $is_directory will be used
+  *                               unquestioningly.  When TRUE, the $path
+  *                               will be inspected to try to determine
+  *                               if it is a file, directory, etc.
+  *                               (OPTIONAL; default = TRUE)
   * @param boolean $quiet         When TRUE, suppresses any error
   *                               messages and makes this function
   *                               return NULL without doing anything
@@ -780,7 +668,7 @@ function ftp_get_file($FTP, $host, $port, $use_ssl, $passive, $mode, $user,
   */
 function ssb_ftp_delete($FTP, $host, $port, $use_ssl, $passive, $user,
                         $pass, $directory, $target, $is_directory,
-                        $quiet=FALSE)
+                        $autodetect=TRUE, $quiet=FALSE)
 {
 
    $error = '';
@@ -798,16 +686,27 @@ function ssb_ftp_delete($FTP, $host, $port, $use_ssl, $passive, $user,
    $file_info = ftp_path_exists(FALSE, $host, $port, $use_ssl,
                                 $passive, $user, $pass,
                                 $full_path, TRUE, $quiet);
+   if (is_null($file_info)) return NULL;  // pass along error ($quiet must be TRUE)
+
+
+   // if file doesn't even exist in the first place, just bail -
+   // our job is already done for us
+   //
+   if (!$file_info)
+      return TRUE;
 
 
    // if possible, try to determine $is_directory on our own
    //
-   if (!empty($file_info['type']) && $file_info['type'] == 'directory')
-      $is_directory = TRUE;
-   else if (!empty($file_info['type']) && $file_info['type'] == 'file')
-      $is_directory = FALSE;
-   else if (!empty($file_info['type']))
-      $is_directory = FALSE;  // for now, we'll assume we can delete other types too but can throw an error here if need be
+   if ($autodetect)
+   {
+      if (!empty($file_info['type']) && $file_info['type'] == 'directory')
+         $is_directory = TRUE;
+      else if (!empty($file_info['type']) && $file_info['type'] == 'file')
+         $is_directory = FALSE;
+      else if (!empty($file_info['type']))
+         $is_directory = FALSE;  // for now, we'll assume we can delete other types too but can throw an error here if need be
+   }
 
 
    if ($is_directory) $delete_function = 'ftp_rmdir';
@@ -861,19 +760,6 @@ function ssb_ftp_delete($FTP, $host, $port, $use_ssl, $passive, $user,
    $close_connection = FALSE;
 
 
-   // check if the file exists in the first place...
-   // if not, just bail - our job is already done for us
-   //
-   if (empty($error)
-    && !ftp_path_exists($FTP, $host, $port, $use_ssl, $passive, $user,
-                        $pass, $full_path, FALSE, $quiet))
-   {
-      if ($close_connection) ftp_close($FTP);
-      else @ftp_chdir($cwd);
-      return TRUE;
-   }
-
-
    // now, actually delete file/directory
    //
    if (empty($error)
@@ -900,6 +786,7 @@ function ssb_ftp_delete($FTP, $host, $port, $use_ssl, $passive, $user,
       sq_change_text_domain('squirrelmail');
    }
 
+
    // success
    //
    if (empty($error))
@@ -1086,7 +973,7 @@ function ftp_get_directory_listing($FTP, $host, $port, $use_ssl, $passive,
 
 
                case 'Windows_NT':
-//TODO: How do windows servers behave? for now, just try and hope that NT gives same format as *nix...  for testing, maybe can try an open ftp server perhaps unless I can find other testers
+//TODO: this is untested
                   $parsed_file = parse_unix_style_ftp_raw_directory_listing($file, $quiet);
                   if (!empty($parsed_file))
                      $parsed_file_list[] = $parsed_file;
@@ -1169,9 +1056,10 @@ function ftp_get_directory_listing($FTP, $host, $port, $use_ssl, $passive,
   *               at least contain a "name" key whose value is the file
   *               or directory name.  Other possible attributes are:
   *               "type" ("file", "directory", "link", "pipe", "socket"),
-  *               "owner", "group", "size", "permissions", "date" (date
-  *               last modified), "link_saki" (link destination),
-  *               "numeric_permissions".
+  *               "owner", "group", "size", "permissions", "date_modified"
+  *               (date last modified), "date_accessed" (date last accessed),
+  *               "date_changed" (date last changed), "link_saki" (link
+  *               destination), "numeric_permissions".
   *
   */
 function ftp_path_exists($FTP, $host, $port, $use_ssl, $passive, $user,
@@ -1313,7 +1201,7 @@ function ftp_path_exists($FTP, $host, $port, $use_ssl, $passive, $user,
 
 
 /**   
-  * Parses a unix-style file listing
+  * Parses a (space-delimited) unix-style file listing
   *      
   * If some fatal error occurs due to misconfiguration or
   * backend failure, an error message is printed and execution
@@ -1333,16 +1221,17 @@ function ftp_path_exists($FTP, $host, $port, $use_ssl, $passive, $user,
   *               a "name" key whose value is the file or directory name.
   *               Other possible attributes are: "type" ("file", "directory",
   *               "link", "pipe", "socket"), "owner", "group", "size",
-  *               "permissions", "date" (date last modified), "link_saki"
-  *               (link destination), "numeric_permissions".  If there is any
-  *               junk or unexpected formatting in the given $file_info
-  *               input, it may be ignored and an empty array returned if
-  *               possible, and errors are only triggered if more serious
-  *               problems occur.
+  *               "permissions", "date_modified" (date last modified),
+  *               "date_accessed" (date last accessed), "date_changed"
+  *               (date last changed), "link_saki" (link destination),
+  *               "numeric_permissions".  If there is any junk or unexpected
+  *               formatting in the given $file_info input, it may be
+  *               ignored and an empty array returned if possible, and
+  *               errors are only triggered if more serious problems occur.
   *            
   */           
 function parse_unix_style_ftp_raw_directory_listing($file_info, $quiet=FALSE)
-{           
+{
 
    $error = '';
 
@@ -1353,15 +1242,23 @@ function parse_unix_style_ftp_raw_directory_listing($file_info, $quiet=FALSE)
    if (strtolower($attributes[0]) == 'total' || sizeof($attributes) != 9)
       return array();
 
-   $file_attributes['permissions'] = $attributes[0];
-   $file_attributes['number']      = $attributes[1];
-   $file_attributes['owner']       = $attributes[2];
-   $file_attributes['group']       = $attributes[3];
-   $file_attributes['size']        = $attributes[4];
-   $file_attributes['date']        = strtotime($attributes[5] . ' '
-                                             . $attributes[6] . ' '
-                                             . $attributes[7]);
-   $file_attributes['name']        = $attributes[8];
+   $file_attributes['permissions']     = $attributes[0];
+   $file_attributes['number_of_links'] = $attributes[1];
+   $file_attributes['size']            = $attributes[4];
+   $file_attributes['date_modified']   = strtotime($attributes[5] . ' '
+                                                 . $attributes[6] . ' '
+                                                 . $attributes[7]);
+   $file_attributes['name']            = $attributes[8];
+
+   if (is_numeric($attributes[2]))
+      $file_attributes['uid'] = $attributes[2];
+   else
+      $file_attributes['owner'] = $attributes[2];
+
+   if (is_numeric($attributes[3]))
+      $file_attributes['gid'] = $attributes[3];
+   else
+      $file_attributes['group'] = $attributes[3];
 
    if (preg_match('/(.*?)\s*->\s*(.*)/', $file_attributes['name'], $matches))
    {
@@ -1384,10 +1281,10 @@ function parse_unix_style_ftp_raw_directory_listing($file_info, $quiet=FALSE)
 
    // parse_file_permissions is found in functions.php
    //
-   $file_attributes['numeric_permissions'] = parse_permissions($file_attributes['permissions']);
+   $file_attributes['numeric_permissions'] = decoct(parse_permissions($file_attributes['permissions']));
 
-   //$file_attributes['date_string'] = date('Y/m/d H:i', $file_attributes['date']);
-   //$file_attributes['date_string'] = date('M d, Y g:ia', $file_attributes['date']);
+   //$file_attributes['date_string'] = date('Y/m/d H:i', $file_attributes['date_modified']);
+   //$file_attributes['date_string'] = date('M d, Y g:ia', $file_attributes['date_modified']);
 
    return $file_attributes;
             
@@ -1767,7 +1664,7 @@ function ssb_ftp_rename($FTP, $host, $port, $use_ssl, $passive, $user, $pass,
    //
    else if (!($cwd = ftp_pwd($FTP)) || (!empty($directory) && !sq_call_function_suppress_errors('ftp_chdir', array($FTP, $directory))))
    {
-      if ($quiet) return TRUE;
+      if ($quiet) return NULL;
       sq_change_text_domain('server_settings_backend');
 
       // show connect details when in debug mode
