@@ -2,7 +2,7 @@
 
 /**
   * SquirrelMail Server Settings Backend Plugin
-  * Copyright (c) 2008-2009 Paul Lesniewski <paul@squirrelmail.org>
+  * Copyright (c) 2008-2010 Paul Lesniewski <paul@squirrelmail.org>
   * Licensed under the GNU GPL. For full terms see the file COPYING.
   *
   * @package plugins
@@ -78,7 +78,7 @@ function test_server_setting($test_value, $backend_info, $quiet=FALSE)
   *                              specifically for testing per the
   *                              test_server_setting() function (or
   *                              similar), and so if there is a
-  *                              different (more efficient kind of
+  *                              different (more efficient) kind of
   *                              lookup for this action, it can be
   *                              enabled.  When given, its value must
   *                              be the value being tested for
@@ -193,8 +193,11 @@ function retrieve_server_setting($backend_info, $quiet=FALSE, $test_value=NULL,
       // ----------------------------------
       //
       case 'local_file':
-//TODO
-         $error = "LOCAL_FILE STORAGE BACKEND FOR SERVER SETTINGS PLUGIN IS AS OF YET UNIMPLEMENTED";
+
+         sq_change_text_domain($original_text_domain);
+         include_once(SM_PATH . 'plugins/server_settings_backend/local_file_backend.php');
+         return retrieve_server_setting_local_file($backend_info, $quiet,
+                                                   $test_value, $return_file_contents);
          break;
 
 
@@ -365,8 +368,10 @@ function put_server_setting($new_value, $backend_info,
       // ----------------------------------
       //
       case 'local_file':
-//TODO
-         $error = "LOCAL_FILE STORAGE BACKEND FOR SERVER SETTINGS PLUGIN IS AS OF YET UNIMPLEMENTED";
+
+         include_once(SM_PATH . 'plugins/server_settings_backend/local_file_backend.php');
+         return put_server_setting_local_file($new_value, $backend_info, $add,
+                                              $remove, $quiet);
          break;
 
 
@@ -416,6 +421,387 @@ function put_server_setting($new_value, $backend_info,
 
 
 
+/**
+  * Executes the "vacation" program on the server such that it
+  * initializes the vacation/autoresponse database.
+  *
+  * This feature is intended to support environments where
+  * the vacation program is installed and needs to be used to
+  * initialize certain data before proper functioning of an
+  * autoresponder.
+  *
+  * If some fatal error occurs due to misconfiguration or
+  * backend failure, an error message is printed and execution
+  * stops immediately, unless $quiet is TRUE.
+  *
+  * @param array   $backend_info An array of configuration values
+  *                              that point to how the "vacation"
+  *                              program can be accessed.
+  * @param string  $interval     When non-null, this is the
+  *                              vacation reply interval to
+  *                              be sent to the initialization
+  *                              command.  This only works on
+  *                              some implementations of the
+  *                              "vacation" program (check
+  *                              "man vacation") and will only
+  *                              work in the local_file set-uid 
+  *                              (SUID) backend if support for
+  *                              it has been compiled into the
+  *                              SUID wrapper. (OPTIONAL; default = NULL)
+  * @param boolean $quiet        When TRUE, suppresses any error
+  *                              messages and makes this function
+  *                              return NULL without doing anything
+  *                              else (OPTIONAL; default = FALSE).
+  *
+  * @return mixed TRUE if the "vacation" program was successfully
+  *               run and the autoresponder database thus initialized;
+  *               NULL if it was not (and $quiet is TRUE).
+  *
+  */
+function vacation_initialize($backend_info, $interval=NULL, $quiet=FALSE)
+{
+
+   include_once(SM_PATH . 'plugins/server_settings_backend/init.php');
+
+   $original_text_domain = sq_change_text_domain('server_settings_backend');
+   $error = '';
+
+
+   if (!is_array($backend_info))
+   {
+      $error = _("Misconfiguration encountered when attempting to initialize autoresponder (malformed backend configuration)");
+   }
+
+
+   // for sanity, make sure all array keys are upper case
+   //
+   else
+   {
+      $temp_backend_info = $backend_info;
+      $backend_info = array();
+      foreach ($temp_backend_info as $key => $value)
+         $backend_info[strtoupper($key)] = $value;
+   }
+
+
+   // do we have an array of lookups?  we can assume so if
+   // there is no VALUE, CUSTOM or BACKEND entry in the array
+   // (at this point, VALUE and CUSTOM are already dealt with)
+   //
+   // when we have multiple lookups, just use the first one (multiple
+   // lookups are only useful for performing multiple save actions)
+   //
+   if (empty($backend_info['BACKEND']))
+      return vacation_initialize(array_shift($backend_info), $quiet);
+
+
+   // normal backend storage... start by getting the backend type
+   //
+   $backend = '';
+   if (!empty($backend_info['BACKEND']))
+      $backend = strtolower($backend_info['BACKEND']);
+   if (empty($error) && empty($backend)
+    || ($backend != 'ftp' && $backend != 'local_file' && $backend != 'sql' && $backend != 'ldap'))
+   {
+      $error = sprintf(_("Misconfiguration encountered when attempting to initialize autoresponder (\"%s\" lookup unknown)"), $backend);
+   }
+
+
+   // only the local_file set-uid (SUID) backend supports executing the
+   // "vacation" program directly... all other backends will have to ask
+   // the web server to do it for them -- this requires special
+   // configuration (see plugin documentation) and may or may not be
+   // feasible in some environments
+   //
+   if (empty($error))
+   {
+
+      // manipulate username, domain and password if needed 
+      //                                    
+      global $username, $domain;
+      list($user, $dom, $pass) = get_credentials($backend_info,
+                                                 $username,
+                                                 $domain, 
+                                                 sqauth_read_password(),
+                                                 $quiet);
+
+      // when all three are null, an error occured and $quiet
+      // is turned on, so we can just fake an error
+      //
+      if ($quiet && is_null($user) && is_null($dom) && is_null($pass))
+         $error = 'dummy';
+
+
+      if ($backend == 'local_file' && empty($error))
+      {
+
+         if (!empty($backend_info['ACCESS_TYPE']) && is_array($backend_info['ACCESS_TYPE']))
+            $access_type = retrieve_server_setting($backend_info['ACCESS_TYPE'], $quiet);
+         else
+            $error = _("Misconfiguration encountered when attempting to initialize autoresponder (file access method is misconfigured)");
+         if (!empty($backend_info['SUID_DEBUG_OUTPUT']) && is_array($backend_info['SUID_DEBUG_OUTPUT']))
+            $suid_debug_file = retrieve_server_setting($backend_info['SUID_DEBUG_OUTPUT'], $quiet);
+         else
+            $suid_debug_file = '';
+
+
+         if ($access_type == 'SUID' && empty($error))
+         {
+
+            if (!empty($backend_info['SUID_LOCATION']) && is_array($backend_info['SUID_LOCATION']))
+               $suid_binary = retrieve_server_setting($backend_info['SUID_LOCATION'], $quiet);
+            else
+               $error = _("Misconfiguration encountered when attempting to initialize autoresponder (SUID_LOCATION is misconfigured)");
+            if (!empty($backend_info['MASTER_SECURITY_CHECK']) && is_array($backend_info['MASTER_SECURITY_CHECK']))
+               $master_security_check = retrieve_server_setting($backend_info['MASTER_SECURITY_CHECK'], $quiet);
+            else
+               $master_security_check = FALSE;
+            if (!empty($backend_info['MASTER_USERNAME']) && is_array($backend_info['MASTER_USERNAME']))
+               $master_username = retrieve_server_setting($backend_info['MASTER_USERNAME'], $quiet);
+            else
+               $master_username = NULL;
+
+
+            // if using "master" security check, grab needed values from environment
+            //
+            if ($master_security_check)
+            {
+               if (!sqGetGlobalVar('SQUIRRELMAIL_MASTER_PASSWORD', $master_password, SQ_SERVER)
+                || $master_password == 'test-password')
+                  $error = _("Misconfiguration encountered when attempting to initialize autoresponder (master password is misconfigured)");
+               if ($master_username
+                && (!sqGetGlobalVar('SQUIRRELMAIL_MASTER_USERNAME', $master_username, SQ_SERVER)
+                 || $master_username == 'test-username'))
+                  $error = _("Misconfiguration encountered when attempting to initialize autoresponder (master username is misconfigured)");
+            }
+            else
+               $master_username = $master_password = NULL;
+
+
+            // now we can ask the SUID backend to exectue the "vacation" program...
+            //
+            // first, set up debugging output if needed
+            //
+            if (!empty($suid_debug_file))
+               $suid_debug = ' 2>>' . $suid_debug_file . ' ';
+            else
+               $suid_debug = '';
+
+
+            // now, actually execute the command
+            //
+            $cmd = $suid_binary
+                 . " 'localhost' " // "server" argument isn't currently used - just use "localhost"
+                 . escapeshellarg($user)
+                 . (is_null($interval)
+                    ? " 'vacation_init' "
+                    : " 'vacation_init_with_interval' " . escapeshellarg($interval))
+                 . $suid_debug;
+
+            if (!($SUID = sq_call_function_suppress_errors('popen', array($cmd, 'w'))))
+            {
+
+               // show command details when in debug mode
+               //
+               sq_change_text_domain('server_settings_backend');
+               global $sm_debug_mode, $server_settings_backend_debug;
+               server_settings_backend_init();
+               if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+               if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+                  $error = sprintf(_("Could not establish pipe to command \"%s\""), $cmd);
+               else
+                  $error = _("Error establishing connection when attempting to initialize autoresponder");
+
+               sq_change_text_domain('squirrelmail');
+
+            }
+            else if (!is_null($master_username) && sq_call_function_suppress_errors('fwrite', array($SUID, $master_username . "\n")) === FALSE)
+            {
+
+               // show command details and password (CAREFUL!) when in debug mode
+               //
+               sq_change_text_domain('server_settings_backend');
+               global $sm_debug_mode, $server_settings_backend_debug;
+               server_settings_backend_init();
+               if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+               if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+                  $error = sprintf(_("Could not send master username \"%s\" to pipe command \"%s\""), $master_username, $cmd);
+               else
+                  $error = _("Unable to send master username when attempting to initialize autoresponder");
+
+               sq_change_text_domain('squirrelmail');
+
+               @unlink($local_file);
+
+            }
+            else if (!is_null($master_password) && sq_call_function_suppress_errors('fwrite', array($SUID, $master_password . "\n")) === FALSE)
+            {
+
+               // show command details and password (CAREFUL!) when in debug mode
+               //
+               sq_change_text_domain('server_settings_backend');
+               global $sm_debug_mode, $server_settings_backend_debug;
+               server_settings_backend_init();
+               if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+               if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+                  $error = sprintf(_("Could not send master password \"%s\" to pipe command \"%s\""), $master_password, $cmd);
+               else
+                  $error = _("Unable to send master password when attempting to initialize autoresponder");
+
+               sq_change_text_domain('squirrelmail');
+
+               @unlink($local_file);
+
+            }
+            else if (sq_call_function_suppress_errors('fwrite', array($SUID, $pass . "\n")) === FALSE)
+            {
+
+               // show command details and password (CAREFUL!) when in debug mode
+               //
+               sq_change_text_domain('server_settings_backend');
+               global $sm_debug_mode, $server_settings_backend_debug;
+               server_settings_backend_init();
+               if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+               if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+                  $error = sprintf(_("Could not send password \"%s\" to pipe command \"%s\""), $pass, $cmd);
+               else
+                  $error = _("Unable to send password when attempting to initialize autoresponder");
+
+               sq_change_text_domain('squirrelmail');
+
+            }
+            else
+            {
+
+               // get command return code
+               //
+               $status = ssb_pclose($SUID);
+
+
+               // no error?  ready to return
+               //
+               if ($status == 0) // 0 is ERR_OK
+                  return TRUE;
+
+
+               // show command details when in debug mode
+               //
+               sq_change_text_domain('server_settings_backend');
+               global $sm_debug_mode, $server_settings_backend_debug;
+               server_settings_backend_init();
+               if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+               if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+                  $error = sprintf(_("Unable to initialize autoresponder using command \"%s\""), $cmd);
+               else
+                  $error = sprintf(_("Unable to initialize autoresponder (%s)"), $status);
+
+               sq_change_text_domain('squirrelmail');
+
+            }
+
+         }
+
+      }
+
+      // all other backends (including Local File with PHP access) get
+      // to rely on the web server to execute the "vacation" program
+      //
+      if (empty($error))
+      {
+
+         // get special "VACATION_PROGRAM" out of backend settings
+         //
+         if (!empty($backend_info['VACATION_PROGRAM']) && is_array($backend_info['VACATION_PROGRAM']))
+            $vacation_program = retrieve_server_setting($backend_info['VACATION_PROGRAM'], $quiet);
+         else
+            $error = _("Misconfiguration encountered when attempting to initialize autoresponder (vacation program is misconfigured)");
+         if (!empty($backend_info['VACATION_INIT_FLAG']) && is_array($backend_info['VACATION_INIT_FLAG']))
+            $vacation_init_flag = retrieve_server_setting($backend_info['VACATION_INIT_FLAG'], $quiet);
+         else
+            $vacation_init_flag = '-I';
+         if (!empty($backend_info['VACATION_INTERVAL_FLAG']) && is_array($backend_info['VACATION_INTERVAL_FLAG']))
+            $vacation_interval_flag = retrieve_server_setting($backend_info['VACATION_INTERVAL_FLAG'], $quiet);
+         else if (!is_null($interval))
+            $error = _("Misconfiguration encountered when attempting to initialize autoresponder (vacation interval flag is misconfigured)");
+         else
+            $vacation_interval_flag = NULL;
+         if (!empty($backend_info['VACATION_USER_ARG']) && is_array($backend_info['VACATION_USER_ARG']))
+            $vacation_user_arg = retrieve_server_setting($backend_info['VACATION_USER_ARG'], $quiet);
+         else
+            $vacation_user_arg = NULL;
+
+
+         // ready to go
+         //
+         if (empty($error))
+         {
+
+            $vacation_program = str_replace(array('%u', '%d'), array(ssb_str_esc($user), ssb_str_esc($dom)), $vacation_program);
+            $vacation_command = $vacation_program . ' '
+                              . escapeshellarg($vacation_init_flag)
+                              . (is_null($interval) || is_null($vacation_interval_flag) ? ''
+                                 : ' ' . escapeshellarg($vacation_interval_flag)
+                                 . ' ' . escapeshellarg($interval))
+                              . (is_null($vacation_user_arg) ? ''
+                                 : ' ' . escapeshellarg(str_replace(array('%u', '%d'), array(ssb_str_esc($user), ssb_str_esc($dom)), $vacation_user_arg)));
+            $result = exec($vacation_command, $output, $retval);
+
+
+            // we expect a returned status of zero to indicate success
+            //
+            if (!$retval) return TRUE;
+
+
+            // otherwise an error occurred
+            //
+            if (!isset($output[0])) $output[0] = '';
+
+
+            // show command details when in debug mode
+            //
+            sq_change_text_domain('server_settings_backend');
+            global $sm_debug_mode, $server_settings_backend_debug;
+            server_settings_backend_init();
+            if ($server_settings_backend_debug) $sm_debug_mode |= SM_DEBUG_MODE_SIMPLE;
+            if ($sm_debug_mode & SM_DEBUG_MODE_SIMPLE)
+               $error = sprintf(_("Error initializing autoresponder (%d): %s"), $retval, $output[0]);
+            else
+               $error = _("Error initializing autoresponder");
+
+            sq_change_text_domain('squirrelmail');
+
+         }
+
+      }
+
+   }
+
+
+   // should not get here without an error message
+   // unless something is very wrong
+   //    
+   if (empty($error))
+      $error = _("Unknown error in Server Settings Backend plugin, function put_server_setting()");
+
+         
+   // spit out error and exit
+   //    
+   if ($quiet) return NULL;
+   sq_change_text_domain('squirrelmail');  // NOT $original_text_domain
+   global $color;
+   $ret = plain_error_message($error, $color);
+   if (check_sm_version (1, 5, 2))
+   {        
+      echo $ret;
+      global $oTemplate;
+      $oTemplate->display('footer.tpl');
+   }  
+   exit;
+
+}
+
+
+
 /**
   * Based on the given configuration and default
   * values, this function possibly modifies the
@@ -637,30 +1023,42 @@ function ssb_str_esc($string)
 
 /**
   * Generates a unique temporary local file name in
-  * the directory of choice.
+  * the SquirrelMail attachment directory.
   *
   * NOTE that the returned file handle will point to
   * an OPEN file in WRITE mode.
   *
-  * @param boolean $quiet     When TRUE, suppresses any error
-  *                           messages and makes this function
-  *                           return NULL without doing anything
-  *                           else (OPTIONAL; default = FALSE).
-  * @param string  $name      The desired file name, which
-  *                           will be used if available (OPTIONAL;
-  *                           default not used).
-  * @param string  $extension The desired file extension, which
-  *                           will be used if available (OPTIONAL;
-  *                           default not used).
+  * @param boolean $quiet       When TRUE, suppresses any error
+  *                             messages and makes this function
+  *                             return NULL without doing anything
+  *                             else (OPTIONAL; default = FALSE).
+  * @param string  $name        The desired file name, which
+  *                             will be used if given and available
+  *                             (OPTIONAL; default not used).
+  * @param string  $extension   The desired file extension, which
+  *                             will be used if given and available
+  *                             (OPTIONAL; default not used).
+  * @param boolean $close_file  Whether or not to close the file handle
+  *                             before returning.  The second element
+  *                             in the return value will be NULL in
+  *                             this case (OPTIONAL; default is NOT to
+  *                             close the open file).
+  * @param boolean $delete_file Whether or not to close the file handle
+  *                             and delete the file before returning.
+  *                             The second element in the return value
+  *                             will be NULL in this case (OPTIONAL;
+  *                             default is NOT to close the open file).
   *
-  * @return mixed A two-element array containing the
-  *               file handle to the (open for writing!)
-  *               temporary file and the full path to
-  *               the file (in that order), or NULL if
+  * @return mixed A two-element array containing the full
+  *               path to the temporary file and either NULL
+  *               (depending on the value of )
+  *               or the file handle to the (open for writing!)
+  *               temporary file (in that order), or NULL if
   *               an error occurred (and $quiet is TRUE).
   *
   */
-function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
+function ssb_get_temp_file($quiet=FALSE, $name='', $extension='',
+                           $close_file=FALSE, $delete_file=FALSE)
 {
 
    // give up after 1000 tries
@@ -668,6 +1066,9 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
    $maximum_tries = 1000;
 
 
+   $error = '';
+
+
    // generate random local filename, use SM attach dir
    //
    global $username, $attachment_dir;
@@ -680,7 +1081,8 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
 
    // using PHP >= 4.3.2 we can be truly atomic here
    //
-   $filemods = check_php_version(4, 3, 2) ? 'x' : 'w';
+   $filemode = check_php_version(4, 3, 2) ? 'x' : 'w';
+
 
    // try pure file name if asked for
    //
@@ -692,7 +1094,7 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
       {
          // try to open for (binary) writing
          //
-         $FILE = @fopen($full_local_file, $filemods);
+         $FILE = @fopen($full_local_file, $filemode);
       }
    }
 
@@ -702,10 +1104,10 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
    if (!is_resource($FILE)) for ($try=0; $try < $maximum_tries; ++$try)
    {
 
-      $local_file = GenerateRandomString(32, '', 7);
+      $random_string = GenerateRandomString(32, '', 7);
       $full_local_file = "$hashed_attachment_dir/"
                        . (!empty($name) ? $name . '-' : '')
-                       . $local_file
+                       . $random_string
                        . (!empty($extension) ? '.' . $extension : '');
 
       // filename collision; try again
@@ -715,7 +1117,7 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
 
       // try to open for (binary) writing
       //
-      $FILE = @fopen($full_local_file, $filemods);
+      $FILE = @fopen($full_local_file, $filemode);
 
       // success!
       //
@@ -734,7 +1136,44 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
       sq_change_text_domain('server_settings_backend');
       $error = _("Could not open temporary local file");
       sq_change_text_domain('squirrelmail');
+   }
 
+
+   // only close file handle if asked to do so
+   //
+   if (empty($error) && ($close_file || $delete_file))
+   {
+      if (!sq_call_function_suppress_errors('fclose', array($FILE)))
+      {
+         if ($quiet) return NULL;
+
+         sq_change_text_domain('server_settings_backend');
+         $error = _("Could not close temporary local file");
+         sq_change_text_domain('squirrelmail');
+      }
+      $FILE = NULL;
+   }
+
+
+   // delete file if needed
+   //
+   if (empty($error) && $delete_file)
+   {
+      if (!sq_call_function_suppress_errors('unlink', array($full_local_file)))
+      {
+         if ($quiet) return NULL;
+
+         sq_change_text_domain('server_settings_backend');
+         $error = _("Could not delete temporary local file");
+         sq_change_text_domain('squirrelmail');
+      }
+   }
+
+
+   // spew error and exit if needed
+   //
+   if (!empty($error))
+   {
       global $color;
       $ret = plain_error_message($error, $color);
       if (check_sm_version (1, 5, 2))
@@ -746,7 +1185,7 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
       exit;
    }
 
-   return array($FILE, $full_local_file);
+   return array($full_local_file, $FILE);
 
 }
 
@@ -784,7 +1223,7 @@ function ssb_get_temp_file($quiet=FALSE, $name='', $extension='')
   *               contain each setting value and its offset in the target
   *               file (in that order) if the setting is found more than
   *               once in the file (the caller should check if the return
-  *               value's first array element is an array itslef to
+  *               value's first array element is an array itself to
   *               determine if the result is non-scalar), or NULL if the
   *               setting is not found in the file or if an error occurred
   *               (and $quiet is TRUE).  Note that if $return_file_contents
@@ -805,7 +1244,28 @@ function parse_value_from_file($file_path, $parse_pattern,
    else
    {
       $file_contents = file($file_path);
-      $file_contents = implode('', $file_contents);
+      if (is_array($file_contents))
+         $file_contents = implode('', $file_contents);
+   }
+
+   // problem reading file?
+   //
+   if ($file_contents === FALSE)
+   {
+      if ($quiet) return NULL;
+      else
+      {
+         sq_change_text_domain('squirrelmail');  // NOT $original_text_domain
+         global $color;
+         $ret = plain_error_message(sprintf(_("Unable to read file \"%s\""), $file_path), $color);
+         if (check_sm_version (1, 5, 2))
+         {
+            echo $ret;
+            global $oTemplate;
+            $oTemplate->display('footer.tpl');
+         }
+         exit;
+      }
    }
 
    $ret = parse_value($file_contents, $parse_pattern,
@@ -845,7 +1305,7 @@ function parse_value_from_file($file_path, $parse_pattern,
   *               contain each setting value and its offset in the original
   *               string (in that order) if the setting is found more than
   *               once in the string (the caller should check if the return
-  *               value's first array element is an array itslef to
+  *               value's first array element is an array itself to
   *               determine if the result is non-scalar), or NULL if the
   *               setting is not found in the string or if an error occurred
   *               (and $quiet is TRUE).
@@ -1328,3 +1788,41 @@ function parse_permissions($permissions)
 
 
 
+/**
+  * Wrapper for pclose() that deals intelligently
+  * with return status of the child process.
+  *
+  * @param resource $PROC Handle on the child process being closed
+  *
+  * @return int The return code of the child process or -1 if the
+  *             child was (not normally) terminated.
+  *
+  */
+function ssb_pclose($PROC)
+{
+
+   $status = pclose($PROC);
+
+
+   // we prefer to use the process control functions
+   // to do this right, but if they are not enabled...
+   //
+   if (!function_exists('pcntl_wexitstatus'))
+   {
+      if ($status > 256) $status /= 256;
+      return $status;
+   }
+
+
+   // check for normal exit condition, and if found,
+   // get exit status in portable manner
+   //
+   if (!pcntl_wifexited($status))
+      return -1;
+   else
+      return pcntl_wexitstatus($status);
+
+}
+
+
+
